@@ -14,9 +14,11 @@
 #include "parse-tree.h"
 #include "tools.h"
 #include "unparse.h"
+#include "flang/Common/enum-set.h"
 #include "flang/Common/idioms.h"
 #include "flang/Common/indirection.h"
 #include "flang/Support/Fortran.h"
+#include "llvm/ADT/StringExtras.h"
 #include "llvm/Frontend/OpenMP/OMP.h"
 #include "llvm/Support/raw_ostream.h"
 #include <string>
@@ -35,6 +37,19 @@ public:
       : out_(out), asFortran_{asFortran} {}
 
   static constexpr const char *GetNodeName(const char *) { return "char *"; }
+
+  template <typename T, typename E, size_t B>
+  static std::string GetMemberNames(const common::EnumSet<E, B> &x) {
+    llvm::ListSeparator sep;
+    std::string s;
+    llvm::raw_string_ostream stream(s);
+    x.IterateOverMembers([&](E e) { stream << sep << T::EnumToString(e); });
+    return stream.str();
+  }
+#define NODE_ENUMSET(T, S) \
+  static std::string GetNodeName(const T::S &x) { \
+    return #S " = {"s + GetMemberNames<T>(x) + "}"s; \
+  }
 #define NODE_NAME(T, N) \
   static constexpr const char *GetNodeName(const T &) { return N; }
 #define NODE_ENUM(T, E) \
@@ -49,6 +64,9 @@ public:
   NODE(std, uint64_t)
   NODE_ENUM(common, CUDADataAttr)
   NODE_ENUM(common, CUDASubprogramAttrs)
+  NODE_ENUM(common, ImportKind)
+  NODE_ENUM(common, OmpDependenceKind)
+  NODE_ENUM(common, OmpMemoryOrderType)
   NODE_ENUM(common, OpenACCDeviceType)
   NODE(format, ControlEditDesc)
   NODE(format::ControlEditDesc, Kind)
@@ -205,15 +223,22 @@ public:
   NODE(parser, CompilerDirective)
   NODE(CompilerDirective, AssumeAligned)
   NODE(CompilerDirective, IgnoreTKR)
+  NODE(CompilerDirective, Inline)
+  NODE(CompilerDirective, IVDep)
+  NODE(CompilerDirective, ForceInline)
   NODE(CompilerDirective, LoopCount)
   NODE(CompilerDirective, NameValue)
+  NODE(CompilerDirective, NoInline)
   NODE(CompilerDirective, Unrecognized)
   NODE(CompilerDirective, VectorAlways)
+  NODE_ENUM(CompilerDirective::VectorLength, VectorLength::Kind)
+  NODE(CompilerDirective, VectorLength)
   NODE(CompilerDirective, Unroll)
   NODE(CompilerDirective, UnrollAndJam)
   NODE(CompilerDirective, NoVector)
   NODE(CompilerDirective, NoUnroll)
   NODE(CompilerDirective, NoUnrollAndJam)
+  NODE(CompilerDirective, Prefetch)
   NODE(parser, ComplexLiteralConstant)
   NODE(parser, ComplexPart)
   NODE(parser, ComponentArraySpec)
@@ -383,6 +408,7 @@ public:
   NODE(parser, TeamValue)
   NODE(parser, ImageSelector)
   NODE(parser, ImageSelectorSpec)
+  NODE(ImageSelectorSpec, Notify)
   NODE(ImageSelectorSpec, Stat)
   NODE(ImageSelectorSpec, Team_Number)
   NODE(parser, ImplicitPart)
@@ -482,115 +508,136 @@ public:
   NODE(parser, NullInit)
   NODE(parser, ObjectDecl)
   NODE(parser, OldParameterStmt)
-  NODE(parser, OmpTypeSpecifier)
-  NODE(parser, OmpTypeNameList)
-  NODE(parser, OmpAdjustArgsClause)
-  NODE(OmpAdjustArgsClause, OmpAdjustOp)
-  NODE_ENUM(OmpAdjustArgsClause::OmpAdjustOp, Value)
-  NODE(parser, OmpAppendArgsClause)
-  NODE(OmpAppendArgsClause, OmpAppendOp)
-  NODE(parser, OmpLocator)
-  NODE(parser, OmpLocatorList)
-  NODE(parser, OmpReductionSpecifier)
-  NODE(parser, OmpArgument)
-  NODE(parser, OmpArgumentList)
-  NODE(parser, OmpMetadirectiveDirective)
-  NODE(parser, OmpMatchClause)
-  NODE(parser, OmpOtherwiseClause)
-  NODE(parser, OmpWhenClause)
-  NODE(OmpWhenClause, Modifier)
-  NODE(parser, OmpDirectiveName)
-  NODE(parser, OmpDirectiveSpecification)
-  NODE_ENUM(OmpDirectiveSpecification, Flags)
-  NODE(parser, OmpTraitPropertyName)
-  NODE(parser, OmpTraitScore)
-  NODE(parser, OmpTraitPropertyExtension)
-  NODE(OmpTraitPropertyExtension, Complex)
-  NODE(parser, OmpTraitProperty)
-  NODE(parser, OmpTraitSelectorName)
-  NODE_ENUM(OmpTraitSelectorName, Value)
-  NODE(parser, OmpTraitSelector)
-  NODE(OmpTraitSelector, Properties)
-  NODE(parser, OmpTraitSetSelectorName)
-  NODE_ENUM(OmpTraitSetSelectorName, Value)
-  NODE(parser, OmpTraitSetSelector)
-  NODE(parser, OmpContextSelectorSpecification)
-  NODE(parser, OmpMapper)
-  NODE(parser, OmpMapType)
-  NODE_ENUM(OmpMapType, Value)
-  NODE(parser, OmpMapTypeModifier)
-  NODE_ENUM(OmpMapTypeModifier, Value)
-  NODE(parser, OmpIteratorSpecifier)
-  NODE(parser, OmpIterator)
-  NODE(parser, OmpAbsentClause)
-  NODE(parser, OmpAffinityClause)
-  NODE(OmpAffinityClause, Modifier)
-  NODE(parser, OmpAlignment)
-  NODE(parser, OmpAlignClause)
-  NODE(parser, OmpAlignedClause)
-  NODE(OmpAlignedClause, Modifier)
-  NODE(parser, OmpAtClause)
-  NODE_ENUM(OmpAtClause, ActionTime)
-  NODE_ENUM(OmpSeverityClause, Severity)
-  NODE(parser, OmpBeginBlockDirective)
-  NODE(parser, OmpBeginLoopDirective)
-  NODE(parser, OmpBeginSectionsDirective)
-  NODE(parser, OmpBlockDirective)
+
   static std::string GetNodeName(const llvm::omp::Directive &x) {
     return llvm::Twine("llvm::omp::Directive = ",
         llvm::omp::getOpenMPDirectiveName(x, llvm::omp::FallbackVersion))
         .str();
   }
-  NODE(parser, OmpClause)
-#define GEN_FLANG_DUMP_PARSE_TREE_CLAUSES
-#include "llvm/Frontend/OpenMP/OMP.inc"
-  NODE(parser, OmpClauseList)
+  static std::string GetNodeName(const llvm::omp::Clause &x) {
+    return llvm::Twine(
+        "llvm::omp::Clause = ", llvm::omp::getOpenMPClauseName(x))
+        .str();
+  }
+  NODE(parser, OmpAbsentClause)
+  NODE(parser, OmpAccessGroup)
+  NODE_ENUM(OmpAccessGroup, Value)
+  NODE(parser, OmpAdjustArgsClause)
+  NODE(OmpAdjustArgsClause, OmpAdjustOp)
+  NODE_ENUM(OmpAdjustArgsClause::OmpAdjustOp, Value)
+  NODE(parser, OmpAffinityClause)
+  NODE(OmpAffinityClause, Modifier)
+  NODE(parser, OmpAlignClause)
+  NODE(parser, OmpAlignedClause)
+  NODE(OmpAlignedClause, Modifier)
+  NODE(parser, OmpAlignment)
+  NODE(parser, OmpAlignModifier)
+  NODE(parser, OmpAllocateClause)
+  NODE(OmpAllocateClause, Modifier)
+  NODE(parser, OmpAllocateDirective)
+  NODE(parser, OmpAllocatorComplexModifier)
+  NODE(parser, OmpAllocatorSimpleModifier)
+  NODE(parser, OmpAlwaysModifier)
+  NODE_ENUM(OmpAlwaysModifier, Value)
+  NODE(parser, OmpAppendArgsClause)
+  NODE(OmpAppendArgsClause, OmpAppendOp)
+  NODE(parser, OmpArgument)
+  NODE(parser, OmpArgumentList)
+  NODE(parser, OmpAtClause)
+  NODE_ENUM(OmpAtClause, ActionTime)
+  NODE(parser, OmpAtomicDefaultMemOrderClause)
+  NODE(parser, OmpAttachModifier)
+  NODE_ENUM(OmpAttachModifier, Value)
+  NODE(parser, OmpAutomapModifier)
+  NODE_ENUM(OmpAutomapModifier, Value)
+  NODE(parser, OmpBaseVariantNames)
+  NODE(parser, OmpBeginDirective)
+  NODE(parser, OmpBeginLoopDirective)
+  NODE(parser, OmpBeginSectionsDirective)
+  NODE(parser, OmpBindClause)
+  NODE_ENUM(OmpBindClause, Binding)
+  NODE(parser, OmpBlockConstruct)
   NODE(parser, OmpCancellationConstructTypeClause)
+  NODE(parser, OmpChunkModifier)
+  NODE_ENUM(OmpChunkModifier, Value)
+  NODE(parser, OmpClause)
+  NODE(parser, OmpClauseList)
+  NODE(parser, OmpCloseModifier)
+  NODE_ENUM(OmpCloseModifier, Value)
+  NODE(parser, OmpCombinerClause)
+  NODE(parser, OmpCombinerExpression)
   NODE(parser, OmpContainsClause)
-  NODE(parser, OmpCriticalDirective)
-  NODE(parser, OmpErrorDirective)
-  NODE(parser, OmpNothingDirective)
-  NODE(parser, OmpDeclareTargetSpecifier)
-  NODE(parser, OmpDeclareTargetWithClause)
-  NODE(parser, OmpDeclareTargetWithList)
-  NODE(parser, OmpMapperSpecifier)
+  NODE(parser, OmpContextSelectorSpecification)
+  NODE(parser, OmpDeclareVariantDirective)
   NODE(parser, OmpDefaultClause)
   NODE_ENUM(OmpDefaultClause, DataSharingAttribute)
-  NODE(parser, OmpVariableCategory)
-  NODE_ENUM(OmpVariableCategory, Value)
   NODE(parser, OmpDefaultmapClause)
-  NODE_ENUM(OmpDefaultmapClause, ImplicitBehavior)
   NODE(OmpDefaultmapClause, Modifier)
-  NODE(parser, OmpDependenceType)
-  NODE_ENUM(OmpDependenceType, Value)
-  NODE(parser, OmpTaskDependenceType)
-  NODE_ENUM(OmpTaskDependenceType, Value)
-  NODE(parser, OmpIndirectClause)
-  NODE(parser, OmpIterationOffset)
-  NODE(parser, OmpIteration)
-  NODE(parser, OmpIterationVector)
-  NODE(parser, OmpDoacross)
-  NODE(OmpDoacross, Sink)
-  NODE(OmpDoacross, Source)
+  NODE_ENUM(OmpDefaultmapClause, ImplicitBehavior)
+  NODE(parser, OmpDeleteModifier)
+  NODE_ENUM(OmpDeleteModifier, Value)
   NODE(parser, OmpDependClause)
   NODE(OmpDependClause, TaskDep)
   NODE(OmpDependClause::TaskDep, Modifier)
-  NODE(parser, OmpDetachClause)
-  NODE(parser, OmpDoacrossClause)
+  NODE(parser, OmpDependenceType)
+  NODE_ENUM(OmpDependenceType, Value)
+  NODE(parser, OmpDepinfoModifier)
   NODE(parser, OmpDestroyClause)
-  NODE(parser, OmpEndBlockDirective)
-  NODE(parser, OmpEndCriticalDirective)
+  NODE(parser, OmpDetachClause)
+  NODE(parser, OmpDeviceClause)
+  NODE(OmpDeviceClause, Modifier)
+  NODE(parser, OmpDeviceModifier)
+  NODE_ENUM(OmpDeviceModifier, Value)
+  NODE(parser, OmpDeviceSafesyncClause)
+  NODE(parser, OmpDeviceTypeClause)
+  NODE_ENUM(OmpDeviceTypeClause, DeviceTypeDescription)
+  NODE(parser, OmpDimsModifier)
+  NODE(parser, OmpDirectiveName)
+  NODE(parser, OmpDirectiveSpecification)
+  NODE_ENUM(OmpDirectiveSpecification, Flag)
+  NODE_ENUMSET(OmpDirectiveSpecification, Flags)
+  NODE(parser, OmpDoacross)
+  NODE(OmpDoacross, Sink)
+  NODE(OmpDoacross, Source)
+  NODE(parser, OmpDoacrossClause)
+  NODE(parser, OmpDynamicAllocatorsClause)
+  NODE(parser, OmpDynGroupprivateClause)
+  NODE(OmpDynGroupprivateClause, Modifier)
+  NODE(parser, OmpEndDirective)
   NODE(parser, OmpEndLoopDirective)
   NODE(parser, OmpEndSectionsDirective)
-  NODE(parser, OmpFailClause)
-  NODE(parser, OmpFromClause)
-  NODE(OmpFromClause, Modifier)
+  NODE(parser, OmpEnterClause)
+  NODE(OmpEnterClause, Modifier)
+  NODE(parser, OmpErrorDirective)
   NODE(parser, OmpExpectation)
   NODE_ENUM(OmpExpectation, Value)
+  NODE(parser, OmpFailClause)
+  NODE(parser, OmpFallbackModifier)
+  NODE_ENUM(OmpFallbackModifier, Value)
+  NODE(parser, OmpFromClause)
+  NODE(OmpFromClause, Modifier)
+  NODE(parser, OmpGrainsizeClause)
+  NODE(OmpGrainsizeClause, Modifier)
+  NODE(parser, OmpGraphIdClause)
+  NODE(parser, OmpGraphResetClause)
   NODE(parser, OmpHintClause)
   NODE(parser, OmpHoldsClause)
   NODE(parser, OmpIfClause)
   NODE(OmpIfClause, Modifier)
+  NODE(parser, OmpIndirectClause)
+  NODE(parser, OmpInitClause)
+  NODE(OmpInitClause, Modifier)
+  NODE(parser, OmpInitializerClause)
+  NODE(parser, OmpInitializerExpression)
+  NODE(parser, OmpInReductionClause)
+  NODE(OmpInReductionClause, Modifier)
+  NODE(parser, OmpInteropType)
+  NODE_ENUM(OmpInteropType, Value)
+  NODE(parser, OmpIteration)
+  NODE(parser, OmpIterationOffset)
+  NODE(parser, OmpIterationVector)
+  NODE(parser, OmpIterator)
+  NODE(parser, OmpIteratorSpecifier)
   NODE(parser, OmpLastprivateClause)
   NODE(OmpLastprivateClause, Modifier)
   NODE(parser, OmpLastprivateModifier)
@@ -599,80 +646,113 @@ public:
   NODE(OmpLinearClause, Modifier)
   NODE(parser, OmpLinearModifier)
   NODE_ENUM(OmpLinearModifier, Value)
-  NODE(parser, OmpStepComplexModifier)
-  NODE(parser, OmpStepSimpleModifier)
-  NODE(parser, OmpLoopDirective)
+  NODE(parser, OmpLocator)
+  NODE(parser, OmpLocatorList)
+  NODE(parser, OmpLooprangeClause)
+  NODE(parser, OmpLowerBound)
   NODE(parser, OmpMapClause)
-  NODE(parser, OmpMessageClause)
   NODE(OmpMapClause, Modifier)
-  static std::string GetNodeName(const llvm::omp::Clause &x) {
-    return llvm::Twine(
-        "llvm::omp::Clause = ", llvm::omp::getOpenMPClauseName(x))
-        .str();
-  }
-  NODE(parser, OmpObject)
-  NODE(parser, OmpObjectList)
+  NODE(parser, OmpMapper)
+  NODE(parser, OmpMapperSpecifier)
+  NODE(parser, OmpMapType)
+  NODE_ENUM(OmpMapType, Value)
+  NODE(parser, OmpMapTypeModifier)
+  NODE_ENUM(OmpMapTypeModifier, Value)
+  NODE(parser, OmpMatchClause)
+  NODE(parser, OmpMessageClause)
+  NODE(parser, OmpMetadirectiveDirective)
   NODE(parser, OmpNoOpenMPClause)
   NODE(parser, OmpNoOpenMPRoutinesClause)
   NODE(parser, OmpNoParallelismClause)
+  NODE(parser, OmpNothingDirective)
+  NODE(parser, OmpNumTasksClause)
+  NODE(parser, OmpNumTeamsClause)
+  NODE(OmpNumTeamsClause, Modifier)
+  NODE(parser, OmpNumThreadsClause)
+  NODE(OmpNumThreadsClause, Modifier)
+  NODE(OmpNumTasksClause, Modifier)
+  NODE(parser, OmpObject)
+  NODE(OmpObject, Invalid)
+  NODE_ENUM(OmpObject::Invalid, Kind)
+  NODE(parser, OmpObjectList)
   NODE(parser, OmpOrderClause)
   NODE(OmpOrderClause, Modifier)
   NODE_ENUM(OmpOrderClause, Ordering)
+  NODE(parser, OmpOrderingModifier)
+  NODE_ENUM(OmpOrderingModifier, Value)
   NODE(parser, OmpOrderModifier)
   NODE_ENUM(OmpOrderModifier, Value)
-  NODE(parser, OmpGrainsizeClause)
-  NODE(OmpGrainsizeClause, Modifier)
+  NODE(parser, OmpOtherwiseClause)
+  NODE(parser, OmpPreferenceSelector)
+  NODE(parser, OmpPreferenceSpecification)
+  NODE(parser, OmpPreferType)
   NODE(parser, OmpPrescriptiveness)
   NODE_ENUM(OmpPrescriptiveness, Value)
-  NODE(parser, OmpNumTasksClause)
-  NODE(OmpNumTasksClause, Modifier)
-  NODE(parser, OmpBindClause)
-  NODE_ENUM(OmpBindClause, Binding)
+  NODE(parser, OmpPresentModifier)
+  NODE_ENUM(OmpPresentModifier, Value)
   NODE(parser, OmpProcBindClause)
   NODE_ENUM(OmpProcBindClause, AffinityPolicy)
-  NODE(parser, OmpReductionModifier)
-  NODE_ENUM(OmpReductionModifier, Value)
   NODE(parser, OmpReductionClause)
   NODE(OmpReductionClause, Modifier)
-  NODE(parser, OmpInReductionClause)
-  NODE(OmpInReductionClause, Modifier)
-  NODE(parser, OmpReductionCombiner)
-  NODE(parser, OmpTaskReductionClause)
-  NODE(OmpTaskReductionClause, Modifier)
-  NODE(parser, OmpInitializerProc)
-  NODE(parser, OmpInitializerClause)
   NODE(parser, OmpReductionIdentifier)
-  NODE(parser, OmpAllocateClause)
-  NODE(OmpAllocateClause, Modifier)
-  NODE(parser, OmpAlignModifier)
-  NODE(parser, OmpAllocatorComplexModifier)
-  NODE(parser, OmpAllocatorSimpleModifier)
+  NODE(parser, OmpReductionModifier)
+  NODE_ENUM(OmpReductionModifier, Value)
+  NODE(parser, OmpReductionSpecifier)
+  NODE(parser, OmpRefModifier)
+  NODE_ENUM(OmpRefModifier, Value)
+  NODE(parser, OmpReplayableClause)
+  NODE(parser, OmpReverseOffloadClause)
   NODE(parser, OmpScheduleClause)
   NODE(OmpScheduleClause, Modifier)
   NODE_ENUM(OmpScheduleClause, Kind)
+  NODE(parser, OmpSelfMapsClause)
+  NODE(parser, OmpSelfModifier)
+  NODE_ENUM(OmpSelfModifier, Value)
   NODE(parser, OmpSeverityClause)
-  NODE(parser, OmpDeviceClause)
-  NODE(OmpDeviceClause, Modifier)
-  NODE(parser, OmpDeviceModifier)
-  NODE_ENUM(OmpDeviceModifier, Value)
-  NODE(parser, OmpDeviceTypeClause)
-  NODE_ENUM(OmpDeviceTypeClause, DeviceTypeDescription)
-  NODE(parser, OmpInteropRuntimeIdentifier)
-  NODE(parser, OmpInteropPreference)
-  NODE(parser, OmpInteropType)
-  NODE_ENUM(OmpInteropType, Value)
-  NODE(parser, OmpInitClause)
-  NODE(OmpInitClause, Modifier)
-  NODE(parser, OmpUseClause)
-  NODE(parser, OmpUpdateClause)
-  NODE(parser, OmpChunkModifier)
-  NODE_ENUM(OmpChunkModifier, Value)
-  NODE(parser, OmpOrderingModifier)
-  NODE_ENUM(OmpOrderingModifier, Value)
-  NODE(parser, OmpSectionBlocks)
-  NODE(parser, OmpSectionsDirective)
+  NODE_ENUM(OmpSeverityClause, SevLevel)
+  NODE(parser, OmpStepComplexModifier)
+  NODE(parser, OmpStepSimpleModifier)
+  NODE(parser, OmpStylizedDeclaration)
+  NODE(parser, OmpStylizedExpression)
+  NODE(parser, OmpStylizedInstance)
+  NODE(OmpStylizedInstance, Instance)
+  NODE(parser, OmpTaskDependenceType)
+  NODE(parser, OmpTaskReductionClause)
+  NODE(OmpTaskReductionClause, Modifier)
+  NODE(parser, OmpThreadLimitClause)
+  NODE(OmpThreadLimitClause, Modifier)
+  NODE(parser, OmpThreadsetClause)
+  NODE_ENUM(OmpThreadsetClause, ThreadsetPolicy)
   NODE(parser, OmpToClause)
   NODE(OmpToClause, Modifier)
+  NODE(parser, OmpTraitProperty)
+  NODE(parser, OmpTraitPropertyExtension)
+  NODE(OmpTraitPropertyExtension, Complex)
+  NODE(parser, OmpTraitPropertyName)
+  NODE(parser, OmpTraitScore)
+  NODE(parser, OmpTraitSelector)
+  NODE(OmpTraitSelector, Properties)
+  NODE(parser, OmpTraitSelectorName)
+  NODE_ENUM(OmpTraitSelectorName, Value)
+  NODE(parser, OmpTraitSetSelector)
+  NODE(parser, OmpTraitSetSelectorName)
+  NODE_ENUM(OmpTraitSetSelectorName, Value)
+  NODE(parser, OmpTransparentClause)
+  NODE(parser, OmpTypeName)
+  NODE(parser, OmpTypeNameList)
+  NODE(parser, OmpUnifiedAddressClause)
+  NODE(parser, OmpUnifiedSharedMemoryClause)
+  NODE(parser, OmpUpdateClause)
+  NODE(parser, OmpUseClause)
+  NODE(parser, OmpVariableCategory)
+  NODE_ENUM(OmpVariableCategory, Value)
+  NODE(parser, OmpWhenClause)
+  NODE(OmpWhenClause, Modifier)
+  NODE(parser, OmpxHoldModifier)
+  NODE_ENUM(OmpxHoldModifier, Value)
+#define GEN_FLANG_DUMP_PARSE_TREE_CLAUSES
+#include "llvm/Frontend/OpenMP/OMP.inc"
+
   NODE(parser, Only)
   NODE(parser, OpenACCAtomicConstruct)
   NODE(parser, OpenACCBlockConstruct)
@@ -687,39 +767,35 @@ public:
   NODE(parser, OpenACCStandaloneDeclarativeConstruct)
   NODE(parser, OpenACCStandaloneConstruct)
   NODE(parser, OpenACCWaitConstruct)
+
+  NODE(parser, OpenMPAllocatorsConstruct)
   NODE(parser, OpenMPAssumeConstruct)
-  NODE(parser, OpenMPDeclarativeAssumes)
-  NODE(parser, OmpAssumeDirective)
-  NODE(parser, OmpEndAssumeDirective)
   NODE(parser, OpenMPAtomicConstruct)
-  NODE(parser, OpenMPBlockConstruct)
   NODE(parser, OpenMPCancelConstruct)
   NODE(parser, OpenMPCancellationPointConstruct)
   NODE(parser, OpenMPConstruct)
   NODE(parser, OpenMPCriticalConstruct)
-  NODE(parser, OpenMPDeclarativeAllocate)
+  NODE(parser, OpenMPDeclarativeAssumes)
   NODE(parser, OpenMPDeclarativeConstruct)
-  NODE(parser, OmpDeclareVariantDirective)
+  NODE(parser, OpenMPDeclareMapperConstruct)
   NODE(parser, OpenMPDeclareReductionConstruct)
   NODE(parser, OpenMPDeclareSimdConstruct)
   NODE(parser, OpenMPDeclareTargetConstruct)
-  NODE(parser, OpenMPDeclareMapperConstruct)
-  NODE_ENUM(common, OmpMemoryOrderType)
-  NODE(parser, OmpMemoryOrderClause)
-  NODE(parser, OmpAtomicDefaultMemOrderClause)
   NODE(parser, OpenMPDepobjConstruct)
-  NODE(parser, OpenMPUtilityConstruct)
   NODE(parser, OpenMPDispatchConstruct)
   NODE(parser, OpenMPFlushConstruct)
+  NODE(parser, OpenMPGroupprivate)
+  NODE(parser, OpenMPInvalidDirective)
   NODE(parser, OpenMPLoopConstruct)
-  NODE(parser, OpenMPExecutableAllocate)
-  NODE(parser, OpenMPAllocatorsConstruct)
+  NODE(parser, OpenMPMisplacedEndDirective)
   NODE(parser, OpenMPRequiresConstruct)
-  NODE(parser, OpenMPSimpleStandaloneConstruct)
-  NODE(parser, OpenMPStandaloneConstruct)
   NODE(parser, OpenMPSectionConstruct)
   NODE(parser, OpenMPSectionsConstruct)
+  NODE(parser, OpenMPSimpleStandaloneConstruct)
+  NODE(parser, OpenMPStandaloneConstruct)
   NODE(parser, OpenMPThreadprivate)
+  NODE(parser, OpenMPUtilityConstruct)
+
   NODE(parser, OpenStmt)
   NODE(parser, Optional)
   NODE(parser, OptionalStmt)
