@@ -1,7 +1,7 @@
 ! RUN: %python %S/test_errors.py %s %flang_fc1
 subroutine s1
-  !ERROR: Array 'z' without ALLOCATABLE or POINTER attribute must have explicit shape
-  common x, y(4), z(:)
+  !ERROR: Array 'x' without ALLOCATABLE or POINTER attribute must have explicit shape
+  common x(:), y(4), z
 end
 
 subroutine s2
@@ -28,17 +28,17 @@ subroutine s5
 end
 
 function f6(x) result(r)
-  !ERROR: Dummy argument 'x' may not appear in a COMMON block
-  !ERROR: ALLOCATABLE object 'y' may not appear in a COMMON block
-  common x,y,z
+  !ERROR: ALLOCATABLE object 'y' may not appear in COMMON block //
+  !ERROR: Dummy argument 'x' may not appear in COMMON block //
+  !ERROR: Function result 'r' may not appear in COMMON block //
+  common y,x,z
   allocatable y
-  !ERROR: Function result 'r' may not appear in a COMMON block
   common r
 end
 
 module m7
-  !ERROR: Variable 'w' with BIND attribute may not appear in a COMMON block
-  !ERROR: Variable 'z' with BIND attribute may not appear in a COMMON block
+  !ERROR: BIND(C) object 'w' may not appear in COMMON block //
+  !ERROR: BIND(C) object 'z' may not appear in COMMON block //
   common w,z
   integer, bind(c) :: z
   integer, bind(c,name="w") :: w
@@ -48,8 +48,8 @@ module m8
   type t
   end type
   class(*), pointer :: x
-  !ERROR: Unlimited polymorphic pointer 'x' may not appear in a COMMON block
-  !ERROR: Unlimited polymorphic pointer 'y' may not appear in a COMMON block
+  !ERROR: Unlimited polymorphic pointer 'x' may not appear in COMMON block //
+  !ERROR: Unlimited polymorphic pointer 'y' may not appear in COMMON block //
   common x, y
   class(*), pointer :: y
 end
@@ -67,7 +67,7 @@ module m10
   type t
   end type
   type(t) :: x
-  !ERROR: Derived type 'x' in COMMON block must have the BIND or SEQUENCE attribute
+  !ERROR: Object 'x' whose derived type 't' is neither SEQUENCE nor BIND(C) may not appear in COMMON block //
   common x
 end
 
@@ -82,7 +82,7 @@ module m11
     integer:: c
   end type
   type(t2) :: x2
-  !ERROR: Derived type variable 'x2' may not appear in a COMMON block due to ALLOCATABLE component
+  !ERROR: COMMON block /c2/ may not have the member 'x2' whose derived type 't2' has a component '%b%a' that is ALLOCATABLE or has default initialization
   common /c2/ x2
 end
 
@@ -97,7 +97,7 @@ module m12
     integer:: c
   end type
   type(t2) :: x2
-  !ERROR: Derived type variable 'x2' may not appear in a COMMON block due to component with default initialization
+  !ERROR: COMMON block /c3/ may not have the member 'x2' whose derived type 't2' has a component '%b%a' that is ALLOCATABLE or has default initialization
   common /c3/ x2
 end
 
@@ -112,3 +112,21 @@ subroutine s14
   !ERROR: 'c' appears as a COMMON block in a BIND statement but not in a COMMON statement
   bind(c) :: /c/
 end
+
+module m15
+  interface
+    subroutine sub
+    end subroutine
+  end interface
+  type t1
+    sequence
+    procedure(sub), pointer, nopass :: pp => sub
+  end type
+  type t2
+    sequence
+    type(t1) :: a
+  end type
+  type(t2) :: x2
+  !ERROR: COMMON block /c4/ may not have the member 'x2' whose derived type 't2' has a component '%a%pp' that is ALLOCATABLE or has default initialization
+  common /c4/ x2
+end
