@@ -50,11 +50,16 @@ contains
   subroutine s(x)
     class(t) :: x
   end
+  subroutine test
+    type(t) x
+    !ERROR: Dummy argument 'x=' (#1) is not OPTIONAL and is not associated with an actual argument in this procedure reference
+    call x%p
+  end
 end
 
 module m4
   type :: t
-    !ERROR: Passed-object dummy argument 'x' of procedure 'a' may not have the POINTER attribute
+    !ERROR: Passed-object dummy argument 'x' of procedure 'a' may not have the POINTER attribute unless INTENT(IN)
     procedure(s1), pointer :: a
     !ERROR: Passed-object dummy argument 'x' of procedure 'b' may not have the ALLOCATABLE attribute
     procedure(s2), pointer, pass(x) :: b
@@ -69,7 +74,7 @@ contains
   end
   subroutine s2(w, x)
     real :: x
-    !ERROR: The type of 'x' has already been declared
+    !ERROR: The type of 'x' has already been declared as REAL(4)
     class(t), allocatable :: x
   end
   subroutine s3(f)
